import datetime as dt
import os
from pathlib import Path

from anyio import Path

from .cfg import Config
from .pipeline import Pipeline, PipelineManager
from .scheduler import SchedulerManager


def init(name: str, conf_path: str = "conf", pipelines_path: str = "pipelines"):
    if name is None:
        name = Path.cwd().name

    os.makedirs(os.path.join(name, "conf"), exist_ok=True)
    os.makedirs(os.path.join(name, "pipelines"), exist_ok=True)

    cfg = Config(base_dir=name)

    with open(os.path.join(name, "README.md"), "w") as f:
        f.write(
            f"# {name.replace('_', ' ').upper()}\n\n"
            f"**created with FlowerPower**\n\n*{dt.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}*\n\n"
        )

    # cfg.update(
    #     cfg={
    #         "run": None,
    #         "params": None,
    #     },
    #     name="pipeline",

    # )
    # cfg.update(
    #     cfg={
    #         "username": None,
    #         "api_url": "http://localhost:8241",
    #         "ui_url": "http://localhost:8242",
    #         "api_key": None,
    #         "pipeline": None,
    #     },
    #     name="tracker",

    # )
    # cfg.update(
    #     {
    #         "data_path": {"type": "memory"},
    #         "event_broker": {"type": "local"},
    #         "pipeline": None,
    #     },
    #     name="scheduler",

    # )
    cfg.write(pipeline=True, tracker=True, scheduler=True)



# from apscheduler import (
#     Event,
#     JobAcquired,
#     JobAdded,
#     JobRemoved,
#     JobDeserializationFailed,
#     ScheduleAdded,
#     ScheduleDeserializationFailed,
#     ScheduleRemoved,
#     SchedulerStopped,
#     SchedulerStarted,
#     ScheduleUpdated,
#     TaskAdded,
#     TaskRemoved,
#     TaskUpdated,
# )
# from loguru import logger

# def job_added(event: Event):
#     logger.info(event.timestamp)
#     logger.info(event.job_id)
#     logger.info(event.task_id)
#     logger.info(event.schedule_id)



import datetime as dt

ALL_TRIGGER_TYPES = [
    "cron",
    "interval",
    "calendarinterval",
    "date",
]
ALL_TRIGGER_KWARGS = {
    "cron": [
        "year",
        "month",
        "week",
        "day",
        "days_of_week",
        "hour",
        "minute",
        "second",
        "start_time",
        "end_time",
        "timezone",
    ],
    "interval": [
        "weeks",
        "days",
        "hours",
        "minutes",
        "seconds",
        "microseconds",
        "start_time",
        "end_time",
    ],
    "calendarinterval": [
        "weeks",
        "days",
        "hours",
        "minutes",
        "seconds",
        "start_time",
        "end_time",
        "timezone",
    ],
    "date": [
        "start_time",
    ],
}


class Trigger:
    def __init__(
        self,
        type: str,
    ):
        if type not in ALL_TRIGGER_TYPES:
            raise ValueError(
                f"Invalid trigger type: {type}. Valid trigger types are: {ALL_TRIGGER_TYPES}"
            )
        self.trigger_type = type

    def _check_kwargs(self, **kwargs):
        for k, v in kwargs.items():
            if k not in ALL_TRIGGER_KWARGS[self.trigger_type]:
                raise ValueError(
                    f"Invalid argument: {k}. Valid arguments are: {ALL_TRIGGER_KWARGS[self.trigger_type]}"
                )

    def get(self, **kwargs):
        self._check_kwargs(**kwargs)
        if self.trigger_type == "cron":
            return self._get_cron_trigger(**kwargs)
        elif self.trigger_type == "interval":
            return self._get_interval_trigger(**kwargs)
        elif self.trigger_type == "calendarinterval":
            return self._get_calendar_trigger(**kwargs)
        elif self.trigger_type == "date":
            return self._get_date_trigger(**kwargs)

    def _get_cron_trigger(
        self,
        start_time: dt.datetime | None = None,
        end_time: dt.datetime | None = None,
        timezone: str | None = None,
        **kwargs,
    ):
        from apscheduler.triggers.cron import CronTrigger

        crontab = kwargs.pop("crontab", None)

        if crontab is not None:
            return (CronTrigger.from_crontab(crontab), kwargs)
        else:
            return CronTrigger(
                year=kwargs.pop("year", None),
                month=kwargs.pop("month", None),
                week=kwargs.pop("week", None),
                day=kwargs.pop("day", None),
                day_of_week=kwargs.pop("days_of_week", None),
                hour=kwargs.pop("hour", None),
                minute=kwargs.pop("minute", None),
                second=kwargs.pop("second", None),
                start_time=start_time,
                end_time=end_time,
                timezone=timezone,
            )

    def _get_interval_trigger(
        self,
        start_time: dt.datetime | None = dt.datetime.now(),
        end_time: dt.datetime | None = None,
        **kwargs,
    ):
        from apscheduler.triggers.interval import IntervalTrigger

        return IntervalTrigger(
            weeks=kwargs.pop("weeks", 0),
            days=kwargs.pop("days", 0),
            hours=kwargs.pop("hours", 0),
            minutes=kwargs.pop("minutes", 0),
            seconds=kwargs.pop("seconds", 0),
            microseconds=kwargs.pop("microseconds", 0),
            start_time=start_time,
            end_time=end_time,
        )

    def _get_calendar_trigger(
        self,
        start_time: dt.datetime | None = dt.datetime.now(),
        end_time: dt.datetime | None = None,
        timezone: str | None = None,
        **kwargs,
    ):
        from apscheduler.triggers.calendarinterval import CalendarIntervalTrigger

        return CalendarIntervalTrigger(
            weeks=kwargs.pop("weeks", 0),
            days=kwargs.pop("days", 0),
            hours=kwargs.pop("hours", 0),
            minutes=kwargs.pop("minutes", 0),
            seconds=kwargs.pop("seconds", 0),
            start_time=start_time,
            end_time=end_time,
            timezone=timezone,
        )

    def _get_date_trigger(self, start_time: dt.datetime, **kwargs):
        from apscheduler.triggers.date import DateTrigger

        return DateTrigger(run_time=start_time)


def get_trigger(trigger_type: str, **kwargs):
    trigger = Trigger(trigger_type)
    return trigger.get(**kwargs)



from sqlalchemy.engine import Engine

ALL_EVENT_BROKERS = [
    "sqlalchemy",
    "asyncpg",
    "psycopg3",
    "postgresql",
    "mqtt",
    "redis",
    "local",
    "memory",
]


class EventBroker:
    def __init__(
        self,
        type: str,
        uri: str | None = None,
        sqla_engine: Engine | None = None,
        host: str = "localhost",
        port: int = 1883,
        username: str | None = None,
        password: str | None = None,
    ):
        self.type = type
        self.uri = uri
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self._sqla_engine = sqla_engine

        if type not in ALL_EVENT_BROKERS:
            raise ValueError(
                f"Invalid event broker type: {type}. Valid event broker types are: {ALL_EVENT_BROKERS}"
            )

    def _setup_asyncpg_event_broker(self):
        from apscheduler.eventbrokers.asyncpg import AsyncpgEventBroker

        if self._sqla_engine is None:
            self._event_broker = AsyncpgEventBroker.from_dsn(dsn=self.uri)
        else:
            self._event_broker = AsyncpgEventBroker.from_async_sqla_engine(
                engine=self._sqla_engine
            )

    def _setup_mqtt_event_broker(self):
        from apscheduler.eventbrokers.mqtt import MQTTEventBroker

        self._event_broker = MQTTEventBroker(
            self.host, self.port, topic="flowerpower/scheduler"
        )
        if (self.username is not None) and (self.password is not None):
            self._event_broker._client.username_pw_set(
                self.username,
                self.password,
            )

    def _setup_redis_event_broker(self):
        from apscheduler.eventbrokers.redis import RedisEventBroker

        if self.uri is None:
            self.uri = f"redis://{self.host}:{self.port}"
        self._event_broker = RedisEventBroker(self.uri)

    def _setup_local_event_broker(self):
        from apscheduler.eventbrokers.local import LocalEventBroker

        self._event_broker = LocalEventBroker()

    def setup(self):
        if self.type in ["sqlalchemy", "asyncpg", "psycopg3", "postgresql"]:
            self._setup_asyncpg_event_broker()
        elif self.type == "mqtt":
            self._setup_mqtt_event_broker()
        elif self.type == "redis":
            self._setup_redis_event_broker()
        else:
            self._setup_local_event_broker()

    def get(self):
        return self._event_broker


def setup_event_broker(
    type: str,
    uri: str | None = None,
    sqla_engine: Engine | None = None,
    host: str = "localhost",
    port: int = 1883,
    username: str | None = None,
    password: str | None = None,
):
    eb = EventBroker(
        type=type,
        uri=uri,
        sqla_engine=sqla_engine,
        host=host,
        port=port,
        username=username,
        password=password,
    )
    eb.setup()
    return eb.get()



import datetime as dt
import importlib.util
import sys

if importlib.util.find_spec("apscheduler"):
    from apscheduler import Scheduler, current_scheduler
    from apscheduler.executors.async_ import AsyncJobExecutor
    from apscheduler.executors.subprocess import ProcessPoolJobExecutor
    from apscheduler.executors.thread import ThreadPoolJobExecutor

    from .helpers import monkey

    monkey.patch_pickle()

else:
    raise ImportError(
        "APScheduler is not installed. Please install it using `pip install"
        "'apscheduler>4.0.0a1'`, 'conda install apscheduler4' or `pip install flowerpower[scheduler]`"
    )

import os
import uuid
from typing import Any

from loguru import logger

from .cfg import Config
from .helpers.datastore import setup_data_store
from .helpers.eventbroker import setup_event_broker


class SchedulerManager(Scheduler):
    def __init__(
        self,
        name: str | None = None,
        base_dir: str | None = None,
        **kwargs,
    ):
        """
        Initializes the Scheduler object.

        Args:
            name (str | None, optional): The name of the scheduler. Defaults to None.
            base_dir (str | None, optional): The flowerpower base path. Defaults to None.
            **kwargs: Additional keyword arguments.

        """
        self.name = name or ""

        if base_dir is None:
            base_dir = os.getcwd()

        self._base_dir = base_dir
        self._conf_dir = os.path.join(base_dir, "conf")  # or conf_dir
        self._pipelines_path = os.path.join(base_dir, "pipelines")  # or pipelines_path

        self.cfg = Config(self._base_dir).scheduler
        # self._data_store = None
        # self._event_broker = None
        # self._sqla_engine = None

        self._setup_data_store()
        self._setup_event_broker()
        self._setup_job_executors()

        cleanup_interval = self.cfg.get("cleanup_interval", None) or {
            "unit": "minutes",
            "value": 15,
        }
        super().__init__(
            data_store=self._data_store,
            event_broker=self._event_broker,
            job_executors=self._job_executors,
            identity=self.name,
            logger=logger,
            cleanup_interval=dt.timedelta(
                **{cleanup_interval.unit: cleanup_interval.value}
            ),
            **kwargs,
        )

        sys.path.append(self._pipelines_path)

    def _setup_data_store(self):
        """
        Sets up the data store based on the configuration.

        If the configuration specifies a data store type, it retrieves the data store and SQLA engine
        using the `get_data_store` function. The data store type is obtained from the `type` key in the
        `data_store` section of the configuration. The engine or URI is obtained from the `uri` key in
        the `data_store` section of the configuration.

        Returns:
            None
        """
        if "data_store" in self.cfg:
            if "type" in self.cfg.data_store:
                self._data_store, self._sqla_engine = setup_data_store(
                    type=self.cfg.data_store.type,
                    engine_or_uri=self.cfg.data_store.get("uri", None),
                )

    def _setup_event_broker(self):
        """
        Sets up the event broker based on the configuration settings.

        If the 'event_broker' key is present in the configuration, it checks for the 'type' key.
        If the 'type' key is present, it initializes the event broker using the specified type and other optional
            parameters.
        The optional parameters include 'uri', 'sqla_engine', 'host', 'port', 'username', and 'password'.

        Parameters:
            None

        Returns:
            None
        """
        if "event_broker" in self.cfg:
            if "type" in self.cfg.event_broker:
                self._event_broker = setup_event_broker(
                    type=self.cfg.event_broker.type,
                    uri=self.cfg.event_broker.get("uri", None),
                    sqla_engine=self._sqla_engine,
                    host=self.cfg.event_broker.get("host", "localhost"),
                    port=self.cfg.event_broker.get("port", 1883),
                    username=self.cfg.event_broker.get("username", None),
                    password=self.cfg.event_broker.get("password", None),
                )

    def _setup_job_executors(self):
        self._job_executors = {
            "async": AsyncJobExecutor(),
            "threadpool": ThreadPoolJobExecutor(),
            "processpool": ProcessPoolJobExecutor(),
        }

    def start_worker(
        self,
        background: bool = False,
    ):
        """
        Starts the worker.

        Parameters:
            background (bool): If True, starts the worker in the background. If False, runs the worker until stopped.
        """
        if background:
            self.start_in_background()
        else:
            self.run_until_stopped()

    def stop_worker(self) -> None:
        """
        Stops the worker and closes the exit stack.

        Returns:
            None
        """
        self.stop()
        self._exit_stack.close()

    def remove_all_schedules(self):
        """
        Removes all schedules from the scheduler.

        This method iterates over all schedules in the scheduler and removes them one by one.

        Parameters:
            None

        Returns:
            None
        """
        for sched in self.get_schedules():
            self.remove_schedule(sched.id)


# Wrapper functions for backward compatibility
def get_schedule_manager(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
) -> SchedulerManager:
    """
    Returns a SchedulerManager instance with the specified name and base path.

    Args:
        name (str | None, optional): The name of the scheduler manager. Defaults to None.
        base_dir (str | None, optional): The base path for the scheduler manager. Defaults to None.
        *args: Additional positional arguments to be passed to the SchedulerManager constructor.
        **kwargs: Additional keyword arguments to be passed to the SchedulerManager constructor.

    Returns:
        SchedulerManager: The initialized SchedulerManager instance.
    """
    manager = SchedulerManager(name, base_dir, *args, **kwargs)
    return manager


def get_current_scheduler_manager() -> SchedulerManager | None:
    """
    Returns the current scheduler manager.

    Returns:
        The current scheduler manager if available, otherwise None.
    """
    return current_scheduler.get()


def get_scheduler(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
) -> SchedulerManager:
    """
    Get the scheduler object.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.

    Returns:
        SchedulerManager: The initialized SchedulerManager instance.
    """
    manager = get_schedule_manager(name, base_dir, *args, **kwargs)
    return manager


def start_worker(
    name: str | None = None,
    base_dir: str | None = None,
    background: bool = False,
    *args,
    **kwargs,
) -> SchedulerManager:
    """
    Start the scheduler.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path for the scheduler. Defaults to None.
        background (bool, optional): Whether to start the scheduler in the background. Defaults to False.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.

    Returns:
        SchedulerManager: The scheduler instance.
    """
    # manager = get_schedule_manager(name, base_dir, role="worker", *args, **kwargs)
    with get_schedule_manager(
        name, base_dir, role="worker", *args, **kwargs
    ) as manager:
        manager.start_worker(background)


# def get_current_scheduler() -> Scheduler | None:
#     return SchedulerManager.get_current_scheduler()


def remove_all_schedules(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
):
    """
    Remove all schedules.

    Args:
        name (str | None, optional): The name of the scheduler. Defaults to None.
        base_dir (str | None, optional): The base path. Defaults to None.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.
    """
    manager = get_schedule_manager(name, base_dir, *args, **kwargs)
    manager.remove_all_schedules()


def add_schedule(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
) -> str:
    """
    Add a schedule to the scheduler.

    Args:
        name (str, optional): The name of the schedule. Defaults to None.
        base_dir (str, optional): The base path for the schedule. Defaults to None.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        str: The ID of the added schedule.
    """
    manager = SchedulerManager(name, base_dir, role="scheduler")
    id_ = manager.add_schedule(*args, **kwargs)
    return id_


def add_job(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
) -> uuid.UUID:
    """
    Add a job to the scheduler. Executes the job immediatly and returns the job id (UUID).
    The job result will be stored in the data store for the given `result_expiration_time` and
    can be fetched using the job id (UUID).

    Args:
        name (str | None): The name of the job. Defaults to None.
        base_dir (str | None): The base path for the job. Defaults to None.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        uuid.UUID: The ID of the added job.
    """
    manager = SchedulerManager(name, base_dir, role="scheduler")
    id_ = manager.add_job(*args, **kwargs)
    return id_, manager


def run_job(
    name: str | None = None,
    base_dir: str | None = None,
    *args,
    **kwargs,
) -> Any:
    """
    Run a job using the SchedulerManager. Executes the job immediatly and returns the result of the job.

    Args:
        name (str, optional): The name of the job. Defaults to None.
        base_dir (str, optional): The base path of the job. Defaults to None.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        Any: The SchedulerManager instance.

    """
    manager = SchedulerManager(name, base_dir, role="scheduler")
    manager.run_job(*args, **kwargs)
    return manager



import os
from pathlib import Path

import yaml
from hamilton.function_modifiers import source, value
from munch import Munch, munchify, unmunchify

from .helpers.templates import (
    PIPELINE_TEMPLATE,  # noqa: F401
    SCHEDULER_TEMPLATE,
    TRACKER_TEMPLATE,
)


class BaseConfig:
    def __init__(self, name: str, base_dir: str | None = None) -> None:
        self._base_dir = base_dir or ""
        self._base_dir = str(self._base_dir).rstrip("/")
        if self._base_dir.endswith("conf"):
            self._base_dir = self._base_dir.rstrip("conf").rstrip("/")
        self.name = name
        self._path = os.path.join(self._base_dir, "conf", self.name + ".yml")

    def load(self):
        """
        Load the configuration parameters from a YAML file.

        Args:
            path (str): The path to the YAML file.

        Returns:
            dict: The loaded configuration parameters.
        """
        if Path(self._path).exists():
            with open(self._path) as f:
                self._cfg = munchify(yaml.full_load(f))
        else:
            self._cfg = None

    def write(self) -> None:
        """
        Write the configuration to a YAML file.

        Args:
            cfg (dict | Munch): The configuration to be written.
            name (str): The name of the file.

        Returns:
            None
        """

        with open(self._path, "w") as f:
            f.write(
                eval(f"{self.name.upper()}_TEMPLATE")
                + yaml.dump(unmunchify(self._cfg), sort_keys=False)
                .replace("null", "")
                .replace("{}", "")
            )

    def update(self, cfg: dict | Munch) -> None:
        """
        Update the configuration with the given parameters.

        Args:
            cfg (dict | Munch): The parameters to update the configuration with.
            name (str): The name of the configuration.

        Returns:
            None
        """
        if self._cfg is None:
            self._cfg = munchify(cfg)
        else:
            self._cfg.update(cfg)
        # self._write()


class TrackerConfig(BaseConfig):
    name = "tracker"
    init_cfg = {
        "username": None,
        "api_url": "http://localhost:8241",
        "ui_url": "http://localhost:8242",
        "api_key": None,
        "pipeline": {},
    }

    def __init__(self, base_dir: str | None = None) -> None:
        super().__init__(self.name, base_dir)

        self.load()
        self._cfg = self._cfg or munchify(self.init_cfg)


class SchedulerConfig(BaseConfig):
    name = "scheduler"
    init_cfg = {
        "data_store": {"type": "memory"},
        "event_broker": {"type": "local"},
        "cleanup_interval": {"unit": "minutes", "value": 15},
        "pipeline": {},
    }

    def __init__(self, base_dir: str | None = None) -> None:
        super().__init__(self.name, base_dir)

        self.load()
        self._cfg = self._cfg or munchify(self.init_cfg)


class PipelineConfig(BaseConfig):
    name = "pipeline"
    init_cfg = {
        "run": {},
        # {
        #    "dev": {"inputs": None, "final_vars": None, "with_tracker": False},
        #    "prod": {"inputs": None, "final_vars": None, "with_tracker": True},
        # },
        "params": {},
    }

    def __init__(self, base_dir: str | None = None) -> None:
        super().__init__(self.name, base_dir)

        self.load()
        self._cfg = self._cfg or munchify(self.init_cfg)

    def _to_ht_params(self, d: dict, parent_dict: dict | None = None):
        """
        Recursively converts the values in a dictionary to `source` or `value` objects.

        Args:
            d (dict): The dictionary to convert.
            parent_dict (dict | None): The parent dictionary. Defaults to None.

        Returns:
            dict: The converted dictionary.
        """

        if parent_dict is None:
            parent_dict = d

        for k, v in d.items():
            if isinstance(v, str):
                if v in parent_dict:
                    d[k] = source(v)
                else:
                    d[k] = value(v)
            else:
                d[k] = value(v)
        return d

    def _to_ht_parameterization(self, d: dict) -> dict:
        """
        Convert a dictionary into a parameterization dictionary.

        Args:
            d (dict): The input dictionary.

        Returns:
            dict: The parameterization dictionary.

        """
        return {k: {k: d[k]} for k in d}

    def load(self):
        """
        Load the configuration parameters.

        This method loads the configuration parameters from the specified file path.
        It updates the parameters dictionary with the loaded values and converts
        certain values to a specific format using the `_to_ht_params` and
        `_to_ht_parameterization` methods.

        Returns:
            None
        """
        ...
        super().load()
        if self._cfg is None:
            return

        self._params = self._cfg.params.copy() if self._cfg.params is not None else {}

        for node in self._params:
            if self._params[node] is None:
                continue
            self._params[node].update(
                {
                    k: self._to_ht_params(v)
                    for k, v in self._params[node].items()
                    if v is not None
                }
            )
        self._params.update(
            {
                k: self._to_ht_parameterization(v)
                for k, v in self._params.items()
                if v is not None
            }
        )
        self._params = munchify(self._params)


class Config:
    def __init__(self, base_dir: str | None = None) -> None:
        self._tracker = TrackerConfig(base_dir)
        self._scheduler = SchedulerConfig(base_dir)
        self._pipeline = PipelineConfig(base_dir)

    def update(
        self,
        cfg: dict | Munch,
        name: str,
    ) -> None:
        """
        Update the configuration with the given parameters.

        Args:
            cfg (dict | Munch): The parameters to update the configuration with.
            name (str): The name of the configuration.

        Returns:
            None
        """
        if name == "pipeline":
            self._pipeline.update(cfg)
        if name == "tracker":
            self._tracker.update(cfg)
        if name == "scheduler":
            self._scheduler.update(cfg)

    def write(
        self, pipeline: bool = False, tracker: bool = False, scheduler: bool = False
    ) -> None:
        """
        Write the configuration to file.

        Args:
            pipeline (bool): If True, write the pipeline configuration to file.
            tracker (bool): If True, write the tracker configuration to file.
            scheduler (bool): If True, write the scheduler configuration to file.

        Returns:
            None
        """
        if pipeline:
            self._pipeline.write()
        if tracker:
            self._tracker.write()
        if scheduler:
            self._scheduler.write()

    @property
    def pipeline(self):
        return self._pipeline._cfg

    @property
    def scheduler(self):
        return self._scheduler._cfg

    @property
    def tracker(self):
        return self._tracker._cfg

    @property
    def pipeline_params(self):
        return self._pipeline._params



import datetime as dt
import importlib
import importlib.util
import os
import sys
from typing import Any, Callable
from uuid import UUID

from hamilton import driver
from hamilton_sdk import adapters
from loguru import logger
from munch import unmunchify

from .cfg import Config
from .helpers.templates import PIPELINE_PY_TEMPLATE

if importlib.util.find_spec("apscheduler"):
    from .scheduler import SchedulerManager
else:
    SchedulerManager = None


from .helpers.executor import get_executor
from .helpers.trigger import ALL_TRIGGER_KWARGS, get_trigger


class PipelineManager:
    def __init__(self, base_dir: str | None = None):
        """
        Initializes the Pipeline object.

        Args:
            base_dir (str | None): The flowerpower base path. Defaults to None.

        Returns:
            None
        """
        self._base_dir = base_dir or ""
        self._conf_dir = os.path.join(self._base_dir, "conf")
        self._pipeline_dir = os.path.join(self._base_dir, "pipelines")

        sys.path.append(self._pipeline_dir)

        # self._load_module()
        self._load_config()

    def _load_module(self, name: str):
        """
        Load a module dynamically.

        Args:
            name (str): The name of the module to load.

        Returns:
            None
        """
        if not hasattr(self, "_module"):
            self._module = importlib.import_module(name)
        else:
            self._module = importlib.reload(self._module)

    def _load_config(self):
        """
        Load the configuration file.

        This method loads the configuration file specified by the `_conf_dir` attribute and
        assigns it to the `cfg` attribute.

        Parameters:
        - None

        Returns:
        - None
        """
        self.cfg = Config(base_dir=self._base_dir)

    def reload_module(self, name: str):
        """
        Reloads the specified module.

        Args:
            name (str): The name of the module to reload.
        """
        self._load_module(name)

    def reload_config(self):
        """
        Reloads the configuration by creating a new instance of the Config class with the base path
        set to the specified configuration path.

        Parameters:
        - self: The current instance of the Pipeline class.

        Returns:
        - None
        """
        self.cfg = Config(base_dir=self._base_dir)

    def _get_driver(
        self,
        name: str,
        executor: str | None = None,
        with_tracker: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> tuple[driver.Driver, Callable | None]:
        """
        Get the driver and shutdown function for a given pipeline.

        Args:
            name (str): The name of the pipeline.
            executor (str | None, optional): The executor to use. Defaults to None.
            with_tracker (bool, optional): Whether to use the tracker. Defaults to False.
            reload (bool, optional): Whether to reload the module. Defaults to False.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            max_tasks (int, optional): The maximum number of tasks. Defaults to 20.
            num_cpus (int, optional): The number of CPUs. Defaults to 4.
            project_id (str, optional): The project ID for the tracker. Defaults to None.
            username (str, optional): The username for the tracker. Defaults to None.
            dag_name (str, optional): The DAG name for the tracker. Defaults to None.
            tags (str, optional): The tags for the tracker. Defaults to None.
            api_url (str, optional): The API URL for the tracker. Defaults to None.
            ui_url (str, optional): The UI URL for the tracker. Defaults to None.

        Returns:
            tuple[driver.Driver, Callable | None]: A tuple containing the driver and shutdown function.
        """
        max_tasks = kwargs.pop("max_tasks", 20)
        num_cpus = kwargs.pop("num_cpus", 4)
        executor_, shutdown = get_executor(
            executor or "local", max_tasks=max_tasks, num_cpus=num_cpus
        )
        if reload or not hasattr(self, "_module"):
            self._load_module(name)

        if with_tracker:
            tracker_cfg = self.cfg.tracker.pipeline.get(name, {})
            tracker_kwargs = {
                key: kwargs.pop(key, None) or tracker_cfg.get(key, None)
                for key in [
                    "project_id",
                    "username",
                    "dag_name",
                    "tags",
                    "api_url",
                    "ui_url",
                ]
            }
            project_id = tracker_kwargs.get("project_id", None)

            if project_id is None:
                raise ValueError(
                    "Please provide a project_id if you want to use the tracker"
                )

            tracker = adapters.HamiltonTracker(project_id=project_id, **tracker_kwargs)

            dr = (
                driver.Builder()
                .with_modules(self._module)
                .enable_dynamic_execution(allow_experimental_mode=True)
                .with_adapters(tracker)
                .with_remote_executor(executor_)
                .build()
            )
        else:
            dr = (
                driver.Builder()
                .with_modules(self._module)
                .enable_dynamic_execution(allow_experimental_mode=True)
                .with_remote_executor(executor_)
                .build()
            )

        return dr, shutdown

    def _run(
        self,
        name: str,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Run the pipeline with the given parameters.

        Args:
            name (str): The name of the pipeline.
            environment (str, optional): The environment to run the pipeline in. Defaults to "dev".
            executor (str | None, optional): The executor to use. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker. Defaults to None.
            reload (bool, optional): Whether to reload the pipeline. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The result of executing the pipeline.
        """
        logger.info(f"Starting pipeline {name} in environment {environment}")

        run_params = self.cfg.pipeline.run.get(name)[environment]

        final_vars = final_vars or run_params.get("final_vars", [])
        inputs = {
            **(run_params.get("inputs", {}) or {}),
            **(inputs or {}),
        }  # <-- inputs override and adds to run_params

        kwargs.update(
            {
                arg: eval(arg) or run_params.get(arg, None)
                for arg in ["executor", "with_tracker", "reload"]
            }
        )

        dr, shutdown = self._get_driver(
            name=name,
            **kwargs,
        )

        res = dr.execute(final_vars=final_vars, inputs=unmunchify(inputs))

        logger.success(f"Finished pipeline {name}")

        if shutdown is not None:
            shutdown()

        return res

    def run(
        self,
        name: str,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Run the pipeline.

        Args:
            name (str): The name of the pipeline.
            environment (str, optional): The environment to run the pipeline in. Defaults to "dev".
            executor (str | None, optional): The executor to use. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool, optional): Whether to use a tracker. Defaults to False.
            reload (bool, optional): Whether to reload the pipeline. Defaults to False.
            **kwargs: Additional keyword arguments.
        Returns:
            Any: The result of running the pipeline.
        """
        return self._run(
            name=name,
            environment=environment,
            inputs=inputs,
            final_vars=final_vars,
            executor=executor,
            with_tracker=with_tracker,
            reload=reload,
            **kwargs,
        )

    def run_job(
        self,
        name: str,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        """
        Add a job to run the pipeline with the given parameters to the scheduler.
        Executes the job immediatly and returns the result of the execution.

        Args:
            name (str): The name of the job.
            environment (str, optional): The environment to run the job in. Defaults to "dev".
            executor (str | None, optional): The executor to use for the job. Defaults to None.
            inputs (dict | None, optional): The inputs for the job. Defaults to None.
            final_vars (list | None, optional): The final variables for the job. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
            reload (bool, optional): Whether to reload the job. Defaults to False.
            **kwargs: Additional keyword arguments.

        Returns:
            Any: The result of the job execution.
        """

        with SchedulerManager(
            name=name, base_dir=self._base_dir, role="scheduler"
        ) as sm:
            # if not any([task.id == "run-pipeline" for task in sm.get_tasks()]):
            #    sm.configure_task(func_or_task_id="run-pipeline", func=self._run)
            return sm.run_job(
                self._run,
                args=(
                    name,
                    environment,
                    inputs,
                    final_vars,
                    executor,
                    with_tracker,
                    reload,
                ),
                kwargs=kwargs,
                job_executor=executor
                if executor in ["async", "threadpool", "processpool"]
                else "async",
            )

    def add_job(
        self,
        name: str,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ) -> UUID:
        """
        Add a job to run the pipeline with the given parameters to the scheduler data store.
        Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store
        for the given `result_expiration_time` and can be fetched using the job id (UUID).

        Args:
            name (str): The name of the job.
            environment (str, optional): The environment for the job. Defaults to "dev".
            executor (str | None, optional): The executor for the job. Defaults to None.
            inputs (dict | None, optional): The inputs for the job. Defaults to None.
            final_vars (list | None, optional): The final variables for the job. Defaults to None.
            with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
            reload (bool, optional): Whether to reload the job. Defaults to False.
            result_expiration_time (float | dt.timedelta | None, optional): The result expiration time for the job.
                Defaults to None.
            **kwargs: Additional keyword arguments.

        Returns:
            UUID: The UUID of the added job.
        """
        with SchedulerManager(
            name=name, base_dir=self._base_dir, role="scheduler"
        ) as sm:
            return sm.add_job(
                self._run,
                args=(
                    name,
                    environment,
                    inputs,
                    final_vars,
                    executor,
                    with_tracker,
                    reload,
                ),
                kwargs=kwargs,
                job_executor=executor
                if executor in ["async", "threadpool", "processpool"]
                else "async",
                result_expiration_time=result_expiration_time,
            )

    def schedule(
        self,
        name: str,
        inputs: dict | None = None,
        final_vars: list | None = None,
        environment: str = "dev",
        executor: str | None = None,
        with_tracker: bool | None = None,
        trigger_type: str | None = None,
        paused: bool = False,
        coalesce: str = "latest",
        misfire_grace_time: float | dt.timedelta | None = None,
        max_jitter: float | dt.timedelta | None = None,
        max_running_jobs: int | None = None,
        conflict_policy: str = "do_nothing",
        **kwargs,
    ) -> str:
        """
        Schedule a pipeline for execution.

        Args:
            name (str): The name of the pipeline.
            environment (str, optional): The environment in which the pipeline will run. Defaults to "dev".
            executor (str | None, optional): The executor to use for running the pipeline. Defaults to None.
            trigger_type (str | None, optional): The type of trigger for the pipeline. Defaults to None.
            inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
            final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
            with_tracker (bool | None, optional): Whether to include a tracker for the pipeline. Defaults to None.
            paused (bool, optional): Whether the pipeline should be initially paused. Defaults to False.
            coalesce (str, optional): The coalesce strategy for the pipeline. Defaults to "latest".
            misfire_grace_time (float | dt.timedelta | None, optional): The grace time for misfired jobs.
                Defaults to None.
            max_jitter (float | dt.timedelta | None, optional): The maximum number of seconds to randomly add to the
                scheduled. Defaults to None.
            max_running_jobs (int | None, optional): The maximum number of running jobs for the pipeline.
                Defaults to None.
            conflict_policy (str, optional): The conflict policy for the pipeline. Defaults to "do_nothing".
            **kwargs: Additional keyword arguments for the trigger.

        Returns:
            str: The ID of the scheduled pipeline.

        Raises:
            ValueError: If APScheduler4 is not installed.

        """
        if SchedulerManager is None:
            raise ValueError("APScheduler4 not installed. Please install it first.")

        if "pipeline" in self.cfg.scheduler:
            scheduler_cfg = self.cfg.scheduler.pipeline.get(name, None)  # .copy()
        else:
            scheduler_cfg = {}

        trigger_type = trigger_type or scheduler_cfg.get("trigger_type", None)

        trigger_kwargs = {
            key: kwargs.pop(key, None) or scheduler_cfg.get(key, None)
            for key in ALL_TRIGGER_KWARGS.get(trigger_type, [])
            if key in kwargs or key in scheduler_cfg
        }

        schedule_kwargs = {
            arg: eval(arg) or scheduler_cfg.get(arg, None)
            for arg in [
                "executor",
                "paused",
                "coalesce",
                "misfire_grace_time",
                "max_jitter",
                "max_running_jobs",
                "conflict_policy",
            ]
        }

        with SchedulerManager(
            name=name, base_dir=self._base_dir, role="scheduler"
        ) as sm:
            trigger = get_trigger(trigger_type, **trigger_kwargs)

            id_ = sm.add_schedule(
                self._run,
                trigger=trigger,
                args=(name, environment, inputs, final_vars, executor, with_tracker),
                kwargs=kwargs,
                **schedule_kwargs,
            )
            logger.success(
                f"Added scheduler for {name} in environment {environment} with id {id_}"
            )
            return id_

    def new(
        self,
        name: str,
        overwrite: bool = False,
        params: dict | None = None,
        run: dict | None = None,
        schedule: dict | None = None,
        tracker: dict | None = None,
    ):
        """
        Add a new item to the pipeline.

        Args:
            name (str): The name of the item.
            overwrite (bool, optional): Whether to overwrite an existing item with the same name. Defaults to False.
            params (dict | None, optional): The parameters for the item. Defaults to None.
            run (dict | None, optional): The run configuration for the item. Defaults to None.
            schedule (dict | None, optional): The schedule configuration for the item. Defaults to None.
            tracker (dict | None, optional): The tracker configuration for the item. Defaults to None.
        """
        self.add(name, overwrite, params, run, schedule, tracker)

    def add(
        self,
        name: str,
        overwrite: bool = False,
        params: dict | None = None,
        run: dict | None = None,
        schedule: dict | None = None,
        tracker: dict | None = None,
    ):
        """
        Adds a new pipeline with the given name.

        Args:
            name (str): The name of the pipeline.
            overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
            params (dict | None, optional): The parameters for the pipeline. Defaults to None.
            run (dict | None, optional): The run configuration for the pipeline. Defaults to None.
            schedule (dict | None, optional): The schedule configuration for the pipeline. Defaults to None.
            tracker (dict | None, optional): The tracker configuration for the pipeline. Defaults to None.
        """
        logger.info(f"Creating new pipeline {name}")

        if not os.path.exists(self._conf_dir):
            raise ValueError(
                f"Configuration path {self._conf_dir} does not exist. Please run flowerpower init first."
            )
        if not os.path.exists(self._pipeline_dir):
            raise ValueError(
                f"Pipeline path {self._pipeline_dir} does not exist. Please run flowerpower init first."
            )

        if os.path.exists(f"{self._pipeline_dir}/{name}.py") and not overwrite:
            raise ValueError(
                f"Pipeline {name} already exists. Use `overwrite=True` to overwrite."
            )

        os.makedirs(self._pipeline_dir, exist_ok=True)
        with open(f"{self._pipeline_dir}/{name}.py", "w") as f:
            f.write(
                PIPELINE_PY_TEMPLATE.format(
                    name=name, date=dt.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
                )
            )
        logger.info(f"Created pipeline module {name}.py")

        run = run or {
            "dev": {"final_vars": [], "inputs": {}},
            "prod": {"final_vars": [], "inputs": {}},
        }

        self.cfg.update({"run": {name: run}, "params": {name: params}}, name="pipeline")
        logger.info(f"Updated pipeline configuration {self._conf_dir}/pipeline.yml")
        self.cfg.update({"pipeline": {name: schedule}}, name="scheduler")
        logger.info(f"Updated scheduler configuration {self._conf_dir}/scheduler.yml")
        self.cfg.update({"pipeline": {name: tracker}}, name="tracker")
        logger.info(f"Updated tracker configuration {self._conf_dir}/tracker.yml")
        self.cfg.write(pipeline=True, scheduler=True, tracker=True)

        logger.success(f"Created pipeline {name}")

    def delete(self, name: str, cfg: bool = True, module: bool = False):
        """
        Delete a pipeline.

        Args:
            name (str): The name of the pipeline to delete.
            cfg (bool, optional): Whether to delete the pipeline configuration. Defaults to True.
            module (bool, optional): Whether to delete the pipeline module file. Defaults to False.
        """

        if cfg:
            self.cfg.pipeline.run.pop(name)
            self.cfg.pipeline.params.pop(name)
            self.cfg.scheduler.pipeline.pop(name)
            self.cfg.tracker.pipeline.pop(name)
            self.cfg.write(pipeline=True, scheduler=True, tracker=True)
            logger.info(f"Deleted pipeline config for {name}")

        if module:
            if os.path.exists(f"{self._pipeline_dir}/{name}.py"):
                os.remove(f"{self._pipeline_dir}/{name}.py")
                logger.info(f"Deleted pipeline module {name}.py")

    def show(
        self, name: str, format: str = "png", view: bool = False, reload: bool = False
    ):
        """
        Display the graph of functions for a given name.

        Args:
            name (str): The name of the graph.
            format (str, optional): The format of the graph file. Defaults to "png".
            view (bool, optional): Whether to open the graph file after generating it. Defaults to False.
            reload (bool, optional): Whether to reload the graph data. Defaults to False.

        Returns:
            graph: The generated graph object.
        """
        os.makedirs(os.path.join(self._base_dir, "graphs"), exist_ok=True)
        dr, _ = self._get_driver(
            name=name, executor=None, with_tracker=False, reload=reload
        )
        graph = dr.display_all_functions(
            os.path.join(self._base_dir, "graphs", f"{name}.{format}")
        )
        if view:
            graph.view()
        return graph


class Pipeline(PipelineManager):
    def __init__(self, name: str, base_dir: str | None = None):
        super().__init__(base_dir)
        self.name = name
        self._load_module(name)

    def run(
        self,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        name = self.name
        return super().run(
            name=name,
            environment=environment,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            reload=reload,
            **kwargs,
        )

    def run_job(
        self,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        reload: bool = False,
        **kwargs,
    ) -> Any:
        name = self.name
        return super().run_job(
            name=name,
            environment=environment,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            reload=reload,
            **kwargs,
        )

    def add_job(
        self,
        environment: str = "dev",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool | None = None,
        reload: bool = False,
        result_expiration_time: float | dt.timedelta = 0,
        **kwargs,
    ) -> UUID:
        name = self.name
        return super().add_job(
            name=name,
            environment=environment,
            executor=executor,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            reload=reload,
            result_expiration_time=result_expiration_time,
            **kwargs,
        )

    def schedule(
        self,
        environment: str = "dev",
        type: str = "cron",
        inputs: dict | None = None,
        final_vars: list | None = None,
        executor: str | None = None,
        with_tracker: bool = False,
        paused: bool = False,
        coalesce: str = "latest",
        misfire_grace_time: float | dt.timedelta | None = None,
        max_jitter: float | dt.timedelta | None = None,
        max_running_jobs: int | None = None,
        conflict_policy: str = "do_nothing",
        **kwargs,
    ):
        name = self.name
        return super().schedule(
            name=name,
            environment=environment,
            executor=executor,
            type=type,
            inputs=inputs,
            final_vars=final_vars,
            with_tracker=with_tracker,
            paused=paused,
            coalesce=coalesce,
            misfire_grace_time=misfire_grace_time,
            max_jitter=max_jitter,
            max_running_jobs=max_running_jobs,
            conflict_policy=conflict_policy,
            **kwargs,
        )

    def show(self, format: str = "png", view: bool = False, reload: bool = False):
        return super().show(self.name, format=format, view=view, reload=reload)

    def delete(self, cfg: bool = True, module: bool = False):
        return super().delete(self.name, cfg, module)

    def reload_module(self):
        return super().reload_module(self.name)


def add(
    name: str,
    overwrite: bool = False,
    params: dict | None = None,
    run: dict | None = None,
    schedule: dict | None = None,
    tracker: dict | None = None,
    base_dir: str | None = None,
):
    """
    Add a pipeline to the PipelineManager.

    Args:
        name (str): The name of the pipeline.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        params (dict | None, optional): The parameters for the pipeline. Defaults to None.
        run (dict | None, optional): The run configuration for the pipeline. Defaults to None.
        schedule (dict | None, optional): The schedule configuration for the pipeline. Defaults to None.
        tracker (dict | None, optional): The tracker configuration for the pipeline. Defaults to None.
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
    """
    pm = PipelineManager(base_dir=base_dir)
    pm.add(name, overwrite, params, run, schedule, tracker)


def new(
    name: str,
    overwrite: bool = False,
    params: dict | None = None,
    run: dict | None = None,
    schedule: dict | None = None,
    tracker: dict | None = None,
    base_dir: str | None = None,
):
    """
    Create a new pipeline.

    Args:
        name (str): The name of the pipeline.
        overwrite (bool, optional): Whether to overwrite an existing pipeline with the same name. Defaults to False.
        params (dict | None, optional): The parameters for the pipeline. Defaults to None.
        run (dict | None, optional): The run configuration for the pipeline. Defaults to None.
        schedule (dict | None, optional): The schedule configuration for the pipeline. Defaults to None.
        tracker (dict | None, optional): The tracker configuration for the pipeline. Defaults to None.
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
    """
    pm = PipelineManager(base_dir=base_dir)
    pm.new(name, overwrite, params, run, schedule, tracker)


def run(
    name: str,
    environment: str = "dev",
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool = False,
    base_dir: str | None = None,
    reload: bool = False,
    **kwargs,
) -> Any:
    """
    Run the pipeline with the given parameters.

    Args:
        name (str): The name of the pipeline.
        environment (str, optional): The environment to run the pipeline in. Defaults to "dev".
        executor (str | None, optional): The executor to use. Defaults to None.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        with_tracker (bool, optional): Whether to use the tracker. Defaults to False.
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
        **kwargs: Additional keyword arguments.
    Returns:
        Any: The result of running the pipeline.
    """
    p = Pipeline(name=name, base_dir=base_dir)
    return p.run(
        environment=environment,
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        **kwargs,
    )


def run_job(
    name: str,
    environment: str = "dev",
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool | None = None,
    base_dir: str | None = None,
    reload: bool = False,
    **kwargs,
) -> Any:
    """
    Add a job to run the pipeline with the given parameters to the scheduler.
    Executes the job immediatly and returns the job result.

    Args:
        name (str): The name of the job.
        environment (str, optional): The environment to run the job in. Defaults to "dev".
        executor (str | None, optional): The executor to use for the job. Defaults to None.
        inputs (dict | None, optional): The inputs for the job. Defaults to None.
        final_vars (list | None, optional): The final variables for the job. Defaults to None.
        with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
        base_dir (str | None, optional): The base path for the job. Defaults to None.
        reload (bool, optional): Whether to reload the job. Defaults to False.
        **kwargs: Additional keyword arguments.

    Returns:
        Any: The result of running the job.
    """

    p = Pipeline(name=name, base_dir=base_dir)
    return p.run_job(
        environment=environment,
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        **kwargs,
    )


def add_job(
    name: str,
    environment: str = "dev",
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool | None = None,
    base_dir: str | None = None,
    reload: bool = False,
    result_expiration_time: float | dt.timedelta = 0,
    **kwargs,
) -> UUID:
    """
    Add a job to run the pipeline with the given parameters to the scheduler data store.
    Executes the job immediatly and returns the job id (UUID). The job result will be stored in the data store for the
    given `result_expiration_time` and can be fetched using the job id (UUID).

    Args:
        name (str): The name of the job.
        environment (str, optional): The environment to run the job in. Defaults to "dev".
        executor (str | None, optional): The executor to use for the job. Defaults to None.
        inputs (dict | None, optional): The inputs for the job. Defaults to None.
        final_vars (list | None, optional): The final variables for the job. Defaults to None.
        with_tracker (bool | None, optional): Whether to use a tracker for the job. Defaults to None.
        base_dir (str | None, optional): The base path for the job. Defaults to None.
        reload (bool, optional): Whether to reload the job. Defaults to False.
        result_expiration_time (float | dt.timedelta | None, optional): The expiration time for the job result.
            Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        UUID: The UUID of the added job.
    """
    p = Pipeline(name=name, base_dir=base_dir)
    return p.add_job(
        environment=environment,
        executor=executor,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        reload=reload,
        result_expiration_time=result_expiration_time,
        **kwargs,
    )


def schedule(
    name: str,
    environment: str = "dev",
    type: str = "cron",
    inputs: dict | None = None,
    final_vars: list | None = None,
    executor: str | None = None,
    with_tracker: bool = False,
    paused: bool = False,
    coalesce: str = "latest",
    misfire_grace_time: float | dt.timedelta | None = None,
    max_jitter: float | dt.timedelta | None = None,
    max_running_jobs: int | None = None,
    conflict_policy: str = "do_nothing",
    base_dir: str | None = None,
    **kwargs,
) -> str:
    """
    Schedule a pipeline for execution.

    Args:
        name (str): The name of the pipeline.
        environment (str, optional): The environment in which the pipeline will run. Defaults to "dev".
        executor (str | None, optional): The executor to use for running the pipeline. Defaults to None.
        type (str, optional): The type of schedule. Defaults to "cron".
        auto_start (bool, optional): Whether to automatically start the pipeline. Defaults to True.
        background (bool, optional): Whether to run the pipeline in the background. Defaults to False.
        inputs (dict | None, optional): The inputs for the pipeline. Defaults to None.
        final_vars (list | None, optional): The final variables for the pipeline. Defaults to None.
        with_tracker (bool, optional): Whether to use a tracker for the pipeline. Defaults to False.
        paused (bool, optional): Whether to start the pipeline in a paused state. Defaults to False.
        coalesce (str, optional): The coalesce strategy for the pipeline. Defaults to "latest".
        misfire_grace_time (float | dt.timedelta | None, optional): The grace time for misfired jobs. Defaults to None.
        max_jitter (float | dt.timedelta | None, optional): The maximum number of seconds to randomly add to the
            scheduled. Defaults to None.
        max_running_jobs (int | None, optional): The maximum number of running jobs. Defaults to None.
        conflict_policy (str, optional): The conflict policy for the pipeline. Defaults to "do_nothing".
        base_dir (str | None, optional): The base path for the pipeline. Defaults to None.
        **kwargs: Additional keyword arguments.

    Returns:
        str: The ID of the scheduled pipeline.
    """
    p = Pipeline(name=name, base_dir=base_dir)
    return p.schedule(
        environment=environment,
        executor=executor,
        type=type,
        inputs=inputs,
        final_vars=final_vars,
        with_tracker=with_tracker,
        paused=paused,
        coalesce=coalesce,
        misfire_grace_time=misfire_grace_time,
        max_jitter=max_jitter,
        max_running_jobs=max_running_jobs,
        conflict_policy=conflict_policy,
        **kwargs,
    )


def show(
    name: str,
    format: str = "png",
    view: bool = False,
    base_dir: str | None = None,
    reload: bool = False,
):
    """
    Display the pipeline with the given name.

    Parameters:
        name (str): The name of the pipeline.
        format (str, optional): The format of the displayed pipeline. Defaults to "png".
        view (bool, optional): Whether to display the pipeline. Defaults to False.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        reload (bool, optional): Whether to reload the pipeline. Defaults to False.
    """
    p = Pipeline(name=name, base_dir=base_dir)
    p.show(format=format, view=view, reload=reload)


def delete(name: str, base_dir: str | None = None, module: bool = False):
    """
    Delete a pipeline.

    Args:
        name (str): The name of the pipeline to delete.
        base_dir (str | None, optional): The base path of the pipeline. Defaults to None.
        module (bool, optional): Whether to delete the pipeline module. Defaults to False.
    """
    p = Pipeline(name=name, base_dir=base_dir)
    p.delete(module=module)
